/* ScreetLexer.cpp - Lexer for Screet */
/* Copyright 2002 Steve Folta */
/* See the "License" file */

#include "ScreetLexer.h"


ScreetLexer::ScreetLexer(ScreetString textIn)
	: text(textIn), curCharPtr(textIn.data())
{
	// we always need to have "nextToken", so load it up now
	theNextToken = getNextToken();
	theSecondToken = getNextToken();
}


ScreetLexer::~ScreetLexer()
{
	/***/
}


ScreetToken ScreetLexer::nextToken()
{
	return theNextToken;
}


ScreetToken ScreetLexer::secondToken()
{
	return theSecondToken;
}


void ScreetLexer::consumeToken()
{
	theNextToken = theSecondToken;
	theSecondToken = getNextToken();
}


ScreetToken ScreetLexer::getNextToken()
{
	register const char* charPtr = curCharPtr;
	const char* stopper = text.end();
	char c;

	ScreetToken token(EOFToken, "");

	// skip any whitespace and comments
	while (charPtr < stopper) {
		c = *charPtr;
		if (c == ' ' || c == '\t' || c == '\n' || c == '\r')
			++charPtr;
		else if (c == '/') {
			if (&charPtr[1] < stopper && (charPtr[1] == '*' || charPtr[1] == '/')) {
				++charPtr;
				if (*charPtr == '*') {
					// C-style comment
					++charPtr;
					while (charPtr < stopper) {
						c = *charPtr++;
						if (c == '*') {
							if (charPtr >= stopper)
								break;
							if (*charPtr == '/') {
								++charPtr;
								break;
								}
							}
						}
					}
				else {
					// C++-style comment
					++charPtr;
					while (charPtr < stopper) {
						c = *charPtr++;
						if (c == '\n' || c == '\r')
							break;
						}
					}
				}
			else
				break;
			}
		else if (c == '#') {
			// script-style comment
			++charPtr;
			while (charPtr < stopper) {
				c = *charPtr++;
				if (c == '\n' || c == '\r')
					break;
				}
			}
		// not whitespace or a comment start -- it's the start of the token
		else
			break;
		}

	if (charPtr >= stopper) {
		curCharPtr = charPtr;
		return token;
		}

	const char* tokenStart = charPtr;
	c = *charPtr++;
	switch (c) {
		case '=':
			if (charPtr < stopper && *charPtr == '=') {
				++charPtr;
				token.type = EqualityToken;
				}
			else
				token.type = EqualsToken;
			break;

		case '<':
			token.type = LessThanToken;
			if (charPtr < stopper) {
				if (*charPtr == '=') {
					++charPtr;
					token.type = LessThanOrEqualToken;
					}
				else if (*charPtr == '<') {
					++charPtr;
					token.type = ShiftLeftToken;
					}
				}
			break;

		case '>':
			token.type = GreaterThanToken;
			if (charPtr < stopper) {
				if (*charPtr == '=') {
					++charPtr;
					token.type = GreaterThanOrEqualToken;
					}
				else if (*charPtr == '>') {
					++charPtr;
					token.type = ShiftRightToken;
					}
				}
			break;

		case '&':
			if (charPtr < stopper && *charPtr == '&') {
				++charPtr;
				token.type = LogicalAndToken;
				}
			else
				token.type = AndToken;
			break;

		case '|':
			if (charPtr < stopper && *charPtr == '|') {
				++charPtr;
				token.type = LogicalOrToken;
				}
			else
				token.type = InclusiveOrToken;
			break;

		case '!':
			token.type = NotToken;
			if (charPtr < stopper && *charPtr == '=') {
				++charPtr;
				token.type = InequalityToken;
				}
			break;

		case '0':	case '1':	case '2':	case '3':	case '4':
		case '5':	case '6':	case '7':	case '8':	case '9':
			while (charPtr < stopper && *charPtr >= '0' && *charPtr <= '9')
				++charPtr;
			if (charPtr < stopper && *charPtr == '.') {
				++charPtr;
				while (charPtr < stopper && *charPtr >= '0' && *charPtr <= '9')
					++charPtr;
				}
			token.type = NumberToken;
			break;

		case '+':
			token.type = PlusToken;
			break;
		case '-':
			token.type = MinusToken;
			break;
		case '*':
			token.type = MultiplyToken;
			break;
		case '/':
			token.type = DivideToken;
			break;
		case '%':
			token.type = ModuloToken;
			break;
		case '^':
			token.type = ExclusiveOrToken;
			break;

		case '(':
			token.type = OpenParenToken;
			break;
		case ')':
			token.type = CloseParenToken;
			break;
		case '[':
			token.type = OpenIndexToken;
			break;
		case ']':
			token.type = CloseIndexToken;
			break;
		case '{':
			token.type = OpenBraceToken;
			break;
		case '}':
			token.type = CloseBraceToken;
			break;
		case '.':
			token.type = DotToken;
			break;
		case ';':
			token.type = SemicolonToken;
			break;
		case ',':
			token.type = CommaToken;
			break;
		case '~':
			token.type = TwiddleToken;
			break;

		case '"':
			token.type = StringLiteralToken;
			while (charPtr < stopper) {
				if (*charPtr++ == '"')
					break;
				}
			break;

		case '\'':
			token.type = StringLiteralToken;
			while (charPtr < stopper) {
				if (*charPtr++ == '\'')
					break;
				}
			break;

		case 'b':
			if (ScreetString(charPtr, stopper).startsWith("reak")) {
				charPtr += 4;
				token.type = BreakToken;
				break;
				}
			else
				goto defaultCase;
			break;

		case 'c':
			if (ScreetString(charPtr, stopper).startsWith("ontinue")) {
				charPtr += 7;
				token.type = ContinueToken;
				break;
				}
			else
				goto defaultCase;
			break;

		case 'd':
			if (charPtr < stopper && *charPtr == 'o') {
				++charPtr;
				token.type = DoToken;
				break;
				}
			else
				goto defaultCase;
			break;

		case 'e':
			if (ScreetString(charPtr, stopper).startsWith("lse")) {
				charPtr += 3;
				token.type = ElseToken;
				break;
				}
			else
				goto defaultCase;
			break;

		case 'f':
			if (ScreetString(charPtr, stopper).startsWith("alse")) {
				charPtr += 4;
				token.type = FalseToken;
				break;
				}
			else
				goto defaultCase;
			break;

		case 'i':
			if (charPtr < stopper && *charPtr == 'f') {
				++charPtr;
				token.type = IfToken;
				break;
				}
			else
				goto defaultCase;
			break;

		case 'l':
			if (ScreetString(charPtr, stopper).startsWith("oop")) {
				charPtr += 3;
				token.type = LoopToken;
				break;
				}
			else
				goto defaultCase;
			break;

		case 'n':
			if (charPtr + 1 < stopper && charPtr[0] == 'i' && charPtr[1] == 'l') {
				charPtr += 2;
				token.type = NilToken;
				break;
				}
			else
				goto defaultCase;
			break;

		case 'r':
			if (ScreetString(charPtr, stopper).startsWith("eturn")) {
				charPtr += 5;
				token.type = ReturnToken;
				break;
				}
			else
				goto defaultCase;
			break;

		case 't':
			if (ScreetString(charPtr, stopper).startsWith("his")) {
				charPtr += 3;
				token.type = ThisToken;
				break;
				}
			else if (ScreetString(charPtr, stopper).startsWith("rue")) {
				charPtr += 3;
				token.type = TrueToken;
				break;
				}
			else
				goto defaultCase;
			break;

		case 'w':
			if (ScreetString(charPtr, stopper).startsWith("hile")) {
				charPtr += 4;
				token.type = WhileToken;
				break;
				}
			else
				goto defaultCase;
			break;

		default:
		defaultCase:
			// id
			if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') || c == '_') {
				while (charPtr < stopper) {
					c = *charPtr;
					if ((c >= 'a' && c <= 'z') ||
					    (c >= 'A' && c <= 'Z') ||
					    (c >= '0' && c <= '9') ||
							c == '_') {
						// regular ID character
						++charPtr;
						}
					else if (c == '-' && charPtr + 1 < stopper) {
						// hyphen is part of an ID if surrounded by regular ID characters
						c = charPtr[1];
						if ((c >= 'a' && c <= 'z') ||
					      (c >= 'A' && c <= 'Z') ||
					      (c >= '0' && c <= '9') ||
						    c == '_') {
							// include the hyphen
							++charPtr;
							}
						else
							break;
						}
					else
						break;
					}
				token.type = IDToken;
				}

			// illegal token -- treat it as EOF
			else
				token.type = EOFToken;
			break;
		}

	token.text.setTo(tokenStart, charPtr);
	curCharPtr = charPtr;
	return token;
}


