/* ScreetListener.cpp - Handles Screet expressions on a TCP port. */
/* Copyright 2002 Steve Folta */
/* See the "License" file */

#include "ScreetListener.h"
#include "ScreetParser.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>


// tweex
#define kBufferChunkSize	512


ScreetListener::ScreetListener(ScreetContext* outerContext, int port)
	: screetContext(outerContext)
{
	// create the socket
	listeningSocket = socket(AF_INET, SOCK_STREAM, 0);
	if (listeningSocket == -1)
		return;

	// bind
	struct sockaddr_in address;
	memset(&address, 0, sizeof(address));
	address.sin_family = AF_INET;
	address.sin_port = htons(port);
	address.sin_addr.s_addr = INADDR_ANY;
	int result =
		bind(listeningSocket, (struct sockaddr*) &address, sizeof(struct sockaddr));
	if (result == -1) {
		close(listeningSocket);
		return;
		}

	// listen
	result = ::listen(listeningSocket, 5);
	if (result == -1) {
		close(listeningSocket);
		return;
		}
}


ScreetListener::~ScreetListener()
{
	if (listeningSocket != -1)
		close(listeningSocket);
}


bool ScreetListener::isValid()
{
	return (listeningSocket != -1);
}


void ScreetListener::listen()
{
	while (true)
		handleConnection();
}


void ScreetListener::handleConnection()
{
	// accept a connection
	struct sockaddr_in callerAddress;
	socklen_t callerAddressSize = sizeof(callerAddress);
	int connectionSocket =
		accept(listeningSocket, (struct sockaddr*) &callerAddress, &callerAddressSize);
	if (connectionSocket == -1)
		return;

	bool keepListening = true;
	while (keepListening)
		keepListening = handleRequest(connectionSocket);

	close(connectionSocket);
}


bool ScreetListener::handleRequest(int connectionSocket)
{
	// read from the connection
	long bufferSize = kBufferChunkSize;
	char* buffer = (char*) malloc(bufferSize);
	int bytesLeftInBuffer = bufferSize;
	int totalBytesRead = 0;
	while (true) {
		// read as much as we can
		int bytesRead =
			recv(connectionSocket, &buffer[totalBytesRead], bytesLeftInBuffer, 0);
		if (bytesRead == 0) {
			// the connection was closed, before we received a complete expression
			close(connectionSocket);
			free(buffer);
			return false;
			}
		totalBytesRead += bytesRead;
		bytesLeftInBuffer -= bytesRead;

		// are we done?
		if (requestIsComplete(buffer, totalBytesRead))
			break;

		// did we run out of buffer space?
		if (bytesLeftInBuffer <= 0) {
			// add another chunk
			bufferSize += kBufferChunkSize;
			buffer = (char*) realloc(buffer, bufferSize);
			bytesLeftInBuffer += kBufferChunkSize;
			}
		}

	if (totalBytesRead == 0)
		return false;

	// handle the request
	ScreetString expression(buffer, totalBytesRead);
	ScreetParser parser(expression, &screetContext);
	ScreetString result = parser.parse();
#ifdef TELNET_TEST
	result += "\n";
#endif
	result.c_str();		// make sure there's a null at the end

	// return the result
	int bytesLeftToSend = result.length() + 1;	// + 1 for the null
	const char* toSend = result.data();
	while (bytesLeftToSend > 0) {
		int bytesSent = send(connectionSocket, toSend, bytesLeftToSend, 0);
		if (bytesSent == -1)
			break;

		bytesLeftToSend -= bytesSent;
		toSend += bytesSent;
		}

	free(buffer);
	return true;
}


bool ScreetListener::requestIsComplete(char* buffer, int length)
{
	const char* stopper = buffer + length;
	char* p = buffer;
	while (p < stopper) {
#ifdef TELNET_TEST
		if (*p == '\n') {
			*p = 0;
			return true;
			}
#endif
		if (*p++ == 0)
			return true;
		}

	return false;
}


