/* ScreetParser.cpp -  */
/* Copyright 2002 Steve Folta */
/* See the "License" file */

#include "ScreetParser.h"
#include "ScreetLexer.h"
#include "ScreetMethodCall.h"
#include "ScreetContext.h"
#include "ScreetArgumentList.h"
#ifdef SCREET_WOOSH
#include "ScreetMethodCall.woosh.h"
#include <stdlib.h>
#endif
#include <stdio.h>


ScreetResult::ScreetResult(const ScreetResult& other)
	: type(other.type), string(other.string)
{
	if (type == ObjectType)
		object = other.object;
	else if (type == NumberType)
		number = other.number;
}


ScreetResult::ScreetResult(ScreetString stringIn)
{
	if (stringIn.substr(0, 4) == "##!!") {
		type = ErrorType;
		string = stringIn;
		}
	else if (stringIn.substr(0, 4) == "##0x") {
		type = ObjectType;
#ifdef SCREET_WOOSH
		string = stringIn;
#else
		object = screetObjectForString(stringIn);
#endif
		}
	else if (stringIn.isNumber()) {
		type = NumberType;
		sscanf(stringIn.c_str(), "%lg", &number);
		}
	else {
		type = StringType;
		string = stringIn;
		}
}


ScreetString ScreetResult::asString()
{
	char str[64];

	switch (type) {
		case StringType:
		case ErrorType:
			return string;

		case NumberType:
			{
			sprintf(str, "%.15g", number);
			ScreetString theString(str);
			theString.detach();
			return theString;
			}

		case ObjectType:
			{
#ifdef SCREET_WOOSH
			return string;
#else
			//**** MOVE THIS
			sprintf(str, "##0x%08X", object);
			ScreetString theString(str);
			theString.detach();
			return theString;
#endif
			}
		}

	return ScreetString();
}


bool ScreetResult::conditionIsTrue()
{
	switch (type) {
		case ObjectType:
		case NumberType:
			return true;

		case StringType:
			return !string.isEmpty();

		case ErrorType:
			return false;
		}

	return false;
}


ScreetResult& ScreetResult::operator=(const ScreetResult& other)
{
	type = other.type;
	switch (type) {
		case StringType:
		case ErrorType:
			string = other.string;
			break;

		case NumberType:
			number = other.number;
			break;

		case ObjectType:
			object = other.object;
			break;
		}

	return *this;
}




ScreetParser::ScreetParser(ScreetString text, ScreetContext* contextIn)
	: lexer(NULL), context(contextIn), isEvaluating(true)
{
	lexer = new ScreetLexer(text);
}


ScreetParser::~ScreetParser()
{
	/***/

	delete lexer;
}


ScreetString ScreetParser::parse()
{
	return parseExpression().asString();
}


ScreetResult ScreetParser::parseStatement()
{
	ScreetToken	currentToken = lexer->nextToken();

	switch (currentToken.type) {
		case OpenBraceToken:
			return parseCompoundStatement();
		case IfToken:
			return parseIfStatement();
		case WhileToken:
			return parseWhileStatement();
		case DoToken:
			return parseDoStatement();
		case LoopToken:
			return parseLoopStatement();
		case BreakToken:
			return parseBreakStatement();
		case ContinueToken:
			return parseContinueStatement();
		case ReturnToken:
			return parseReturnStatement();

		case EOFToken:
			return ScreetResult();

		default:
			return parseExpressionStatement();
		}
}


ScreetResult ScreetParser::parseCompoundStatement()
{
	lexer->consumeToken();	// "{"

	ScreetResult result;
	while (true) {
		// are we at the end?
		ScreetTokenType tokenType = lexer->nextToken().type;
		if (tokenType == CloseBraceToken)
			break;
		else if (tokenType == EOFToken)
			return ScreetResult(ErrorType, "##!! The \"{\" is missing from a \"{}\" block. !!##");

		// parse the next statement
		result = parseStatement();
		if (result.type == ErrorType)
			return result;
		}

	lexer->consumeToken();	// "}"

	return result;
}


ScreetResult ScreetParser::parseIfStatement()
{
	lexer->consumeToken();
	bool wasEvaluating = isEvaluating;
	ScreetResult result;

	if (lexer->nextToken().type != OpenParenToken)
		result.error("##!! \"if\" statement is missing the \"(\". !!##");
	else {
		lexer->consumeToken();
		result = parseExpression();
		if (result.type != ErrorType) {
			if (lexer->nextToken().type != CloseParenToken)
				result.error("##!! The \")\" is missing after an \"if\" statement. !!##");
			else {
				lexer->consumeToken();
				bool conditionWasTrue = result.conditionIsTrue();

				// parse the "then" branch
				isEvaluating = (conditionWasTrue ? wasEvaluating : false);
				result = parseStatement();
				if (!isEvaluating)
					result.setTo(StringType, "");

				// check for an "else" branch
				if (result.type != ErrorType && lexer->nextToken().type == ElseToken) {
					lexer->consumeToken();
					isEvaluating = (!conditionWasTrue ? wasEvaluating : false);
					ScreetResult elseResult = parseStatement();
					if (elseResult.type == ErrorType || !conditionWasTrue)
						result = elseResult;
					}
				}
			}
		}

	isEvaluating = wasEvaluating;
	return result;
}


ScreetResult ScreetParser::parseWhileStatement()
{
	/***/
	return ScreetResult(ErrorType, "##!! \"while\" statement is not yet implemented. !!##");
}


ScreetResult ScreetParser::parseDoStatement()
{
	/***/
	return ScreetResult(ErrorType, "##!! \"do\" statement is not yet implemented. !!##");
}


ScreetResult ScreetParser::parseLoopStatement()
{
	/***/
	return ScreetResult(ErrorType, "##!! \"loop\" statement is not yet implemented. !!##");
}


ScreetResult ScreetParser::parseBreakStatement()
{
	/***/
	return ScreetResult(ErrorType, "##!! \"break\" statement is not yet implemented. !!##");
}


ScreetResult ScreetParser::parseContinueStatement()
{
	/***/
	return ScreetResult(ErrorType, "##!! \"continue\" statement is not yet implemented. !!##");
}


ScreetResult ScreetParser::parseReturnStatement()
{
	/***/
	return ScreetResult(ErrorType, "##!! \"return\" statement is not yet implemented. !!##");
}


ScreetResult ScreetParser::parseExpressionStatement()
{
	ScreetResult result = parseExpression();

	if (lexer->nextToken().type != SemicolonToken)
		result.setTo(ErrorType, "##!! A statement is missing the \";\" at the end. !!##");
	else
		lexer->consumeToken();

	return result;
}


ScreetResult ScreetParser::parseExpression()
{
	if (lexer->nextToken().type == OpenBraceToken)
		return parseCompoundStatement();
	else
		return parseAssignmentExpression();
}


ScreetResult ScreetParser::parseAssignmentExpression()
{
	if (lexer->nextToken().type == IDToken &&
	    lexer->secondToken().type == EqualsToken) {
		// assignment
		// get the value
		ScreetString variableName = lexer->nextToken().text;
		lexer->consumeToken();
		lexer->consumeToken();
		ScreetResult value = parseExpression();
		if (value.type == ErrorType)
			return value;

		// set the variable
		// make sure to detach the name and the value -- we don't
		// know for sure that the text we're parsing will stick around
		variableName.detach();
		ScreetString valueStr = value.asString();
		valueStr.detach();
		context->set(variableName, valueStr);
		return value;
		}

	else
		return parseLogicalOrExpression();
}


ScreetResult ScreetParser::parseLogicalOrExpression()
{
	ScreetResult result = parseLogicalAndExpression();

	while (result.type != ErrorType && lexer->nextToken().type == LogicalOrToken) {
		lexer->consumeToken();
		bool wasEvaluating = isEvaluating;

		bool conditionWasTrue = result.conditionIsTrue();
		isEvaluating = (!conditionWasTrue ? wasEvaluating : false);
		ScreetResult rightHandResult = parseLogicalAndExpression();
		if (!conditionWasTrue)
			result = rightHandResult;

		isEvaluating = wasEvaluating;
		}

	return result;
}


ScreetResult ScreetParser::parseLogicalAndExpression()
{
	ScreetResult result = parseInclusiveOrExpression();

	while (result.type != ErrorType && lexer->nextToken().type == LogicalAndToken) {
		lexer->consumeToken();
		bool wasEvaluating = isEvaluating;

		bool conditionWasTrue = result.conditionIsTrue();
		isEvaluating = (conditionWasTrue ? wasEvaluating : false);
		ScreetResult rightHandResult = parseInclusiveOrExpression();
		if (conditionWasTrue)
			result = rightHandResult;

		isEvaluating = wasEvaluating;
		}

	return result;
}


ScreetResult ScreetParser::parseInclusiveOrExpression()
{
	ScreetResult result = parseExclusiveOrExpression();

	while (result.type != ErrorType && lexer->nextToken().type == InclusiveOrToken) {
		lexer->consumeToken();
		ScreetResult arg = parseExclusiveOrExpression();
		if (result.type == NumberType && arg.type == NumberType)
			result.number = ((int) result.number) | ((int) arg.number);
		else if (result.type == ObjectType)
			result = callMethod(result, "|", arg);
		else
			result.error("##!! The '|' operation isn't supported on strings. !!##");
		}

	return result;
}


ScreetResult ScreetParser::parseExclusiveOrExpression()
{
	ScreetResult result = parseAndExpression();

	while (result.type != ErrorType && lexer->nextToken().type == ExclusiveOrToken) {
		lexer->consumeToken();
		ScreetResult arg = parseAndExpression();
		if (result.type == NumberType && arg.type == NumberType)
			result.number = ((int) result.number) ^ ((int) arg.number);
		else if (result.type == ObjectType)
			result = callMethod(result, "^", arg);
		else
			result.error("##!! The '^' operation isn't supported on strings. !!##");
		}

	return result;
}


ScreetResult ScreetParser::parseAndExpression()
{
	ScreetResult result = parseEqualityExpression();

	while (result.type != ErrorType && lexer->nextToken().type == AndToken) {
		lexer->consumeToken();
		ScreetResult arg = parseEqualityExpression();
		if (result.type == NumberType && arg.type == NumberType)
			result.number = ((int) result.number) & ((int) arg.number);
		else if (result.type == ObjectType)
			result = callMethod(result, "&", arg);
		else
			result.error("##!! The '&' operation isn't supported on strings. !!##");
		}

	return result;
}


ScreetResult ScreetParser::parseEqualityExpression()
{
	ScreetResult object = parseAdditiveExpression();

	while (object.type != ErrorType) {
		ScreetTokenType nextToken = lexer->nextToken().type;
		if (nextToken != EqualityToken && nextToken != InequalityToken)
			break;

		lexer->consumeToken();

		// get the argument
		ScreetResult arg = parseAdditiveExpression();
		if (arg.type == ErrorType)
			return arg;

		// perform the operation
		if (object.type == ObjectType) {
			const char* selector = (nextToken == EqualityToken ? "==" : "!=");
			object = callMethod(object, selector, arg);
			}
		else {
			bool result = false;
			if (object.type == NumberType && arg.type == NumberType)
				result = (object.number == arg.number);
			else {
				// treat it as a String
				result = (object.asString() == arg.asString());
				}
			if (nextToken == InequalityToken)
				result = !result;
			object.setTo(StringType, (result ? "true" : ""));
			}
		}

	return object;
}


ScreetResult ScreetParser::parseAdditiveExpression()
{
	ScreetResult object = parseMultiplicativeExpression();

	while (object.type != ErrorType) {
		ScreetTokenType nextToken = lexer->nextToken().type;

		if (nextToken == PlusToken) {
			lexer->consumeToken();

			// get the addend
			ScreetResult arg = parseMultiplicativeExpression();
			if (arg.type == ErrorType)
				return arg;		//** or our own error?

			// perform the addition, according to the type
			if (object.type == NumberType && arg.type == NumberType)
				object.number = object.number + arg.number;
			else if (object.type == ObjectType)
				object = callMethod(object, "+", arg);
			else /* object.type == StringType */
				object.setTo(StringType, object.asString() + arg.asString());
			}

		else if (nextToken == MinusToken) {
			lexer->consumeToken();

			// get the subtrahend
			ScreetResult arg = parseMultiplicativeExpression();
			if (arg.type == ErrorType)
				return arg;

			// perform the subtraction, according to the type
			if (object.type == NumberType && arg.type == NumberType)
				object.number -= arg.number;
			else if (object.type == ObjectType)
				object = callMethod(object, "-", arg);
			else
				object.error("##!! The '-' operation isn't supported on strings. !!##");
			}

		else
			break;
		}

	return object;
}


ScreetResult ScreetParser::parseMultiplicativeExpression()
{
	ScreetResult object = parseUnaryExpression();

	while (object.type != ErrorType) {
		ScreetTokenType nextToken = lexer->nextToken().type;

		if (nextToken == MultiplyToken) {
			lexer->consumeToken();

			// get the multiplier
			ScreetResult arg = parseUnaryExpression();
			if (arg.type == ErrorType)
				return arg;		//** or our own error?

			// perform the multiplication, according to the type
			if (object.type == NumberType && arg.type == NumberType)
				object.number *= arg.number;
			else if (object.type == ObjectType)
				object = callMethod(object, "*", arg);
			else /* object.type == StringType */
				object.error("##!! The '*' operation isn't supported on strings. !!##");
			}

		else if (nextToken == DivideToken) {
			lexer->consumeToken();

			// get the divisor
			ScreetResult arg = parseUnaryExpression();
			if (arg.type == ErrorType)
				return arg;

			// perform the division, according to the type
			if (object.type == NumberType && arg.type == NumberType)
				object.number /= arg.number;
			else if (object.type == ObjectType)
				object = callMethod(object, "/", arg);
			else
				object.error("##!! The '/' operation isn't supported on strings. !!##");
			}

		else if (nextToken == ModuloToken) {
			lexer->consumeToken();

			// get the divisor
			ScreetResult arg = parseUnaryExpression();
			if (arg.type == ErrorType)
				return arg;

			// perform the modulo, according to the type
			if (object.type == NumberType && arg.type == NumberType)
				object.number = ((int) object.number) % ((int) arg.number);
			else if (object.type == ObjectType)
				object = callMethod(object, "%", arg);
			else
				object.error("##!! The '%' operation isn't supported on strings. !!##");
			}

		else
			break;
		}

	return object;
}


ScreetResult ScreetParser::parseUnaryExpression()
{
	switch (lexer->nextToken().type) {
		case NotToken:
			{
			lexer->consumeToken();
			ScreetResult value = parseUnaryExpression();
			if (value.type == ErrorType)
				return value;
			else if (value.type == ObjectType)
				return callMethod(value, "!");
			else
				return ScreetResult(value.asString().isEmpty() ? "true" : "");
			}
			break;

		case PlusToken:
			{
			lexer->consumeToken();
			ScreetResult value = parseUnaryExpression();
			if (value.type == ObjectType)
				return callMethod(value, "unary+");
			else
				return value;
			}
			break;

		case MinusToken:
			{
			lexer->consumeToken();
			ScreetResult value = parseUnaryExpression();
			if (value.type == ObjectType)
				return callMethod(value, "unary-");
			else if (value.type == NumberType)
				return ScreetResult(-value.number);
			else if (value.type == StringType)
				value.error("##!! The unary minus operation is not supported on strings. !!##");
			return value;
			}
			break;

		case TwiddleToken:
			{
			lexer->consumeToken();
			ScreetResult value = parseUnaryExpression();
			if (value.type == ObjectType)
				return callMethod(value, "~");
			else if (value.type == NumberType)
				return ScreetResult(~((int) value.number));
			else if (value.type == StringType)
				value.error("##!! The '~' operation is not supported on strings. !!##");
			return value;
			}
			break;

		default:
			return parsePostfixExpression();
			break;
		}
}


ScreetResult ScreetParser::parsePostfixExpression()
{
	ScreetResult value = parsePrimaryExpression();

	while (value.type != ErrorType) {
		ScreetTokenType nextToken = lexer->nextToken().type;
		if (nextToken == DotToken) {
			// get the method name
			lexer->consumeToken();
			if (lexer->nextToken().type != IDToken) {
				ScreetString errorStr = "##!! A \".\" is followed by \"";
				errorStr += lexer->nextToken().text;
				errorStr += "\", not an identifier. !!##";
				value.error(errorStr);
				break;
				}
			ScreetString methodName = lexer->nextToken().text;
			lexer->consumeToken();

			// get the arguments (if there are any)
			ScreetArgumentList* arguments = NULL;
			if (lexer->nextToken().type == OpenParenToken) {
				lexer->consumeToken();

				// get the arguments
				arguments = parseArguments();
				if (arguments->isError()) {
					ScreetString errorString = arguments->errorString();
					delete arguments;
					value.error(errorString);
					break;
					}

				// make sure there's a ')'
				if (lexer->nextToken().type != CloseParenToken) {
					delete arguments;
					value.error("##!! Unterminated arguments. !!##");
					break;
					}
				lexer->consumeToken();
				}
			else
				arguments = new ScreetArgumentList();

			// see if it's a setter
			if (lexer->nextToken().type == EqualsToken) {
				lexer->consumeToken();

				// get the value to set it to
				ScreetResult setValue = parseExpression();
				if (setValue.type == ErrorType) {
					value = setValue;
					delete arguments;
					break;
					}

				// transform into a setter
				methodName += "-set";
				arguments->addArgument(setValue.asString());
				}

			// perform the method call
			value = callMethod(value, methodName, arguments);
			}

		else if (nextToken == IDToken) {
			// get the function name
			ScreetString methodName = lexer->nextToken().text;
			lexer->consumeToken();

			// get the arguments
			ScreetArgumentList* arguments = parseCommandArguments();
			if (arguments->isError()) {
				ScreetString errorString = arguments->errorString();
				delete arguments;
				value.error(errorString);
				break;
				}

			// perform the method call
			value = callMethod(value, methodName, arguments);
			}

		else if (nextToken == OpenIndexToken) {
			/***/
			value.error("##!! [] is not yet implemented. !!##");
			}

		else
			break;
		}

	return value;
}


ScreetResult ScreetParser::parsePrimaryExpression()
{
	switch (lexer->nextToken().type) {
		case ThisToken:
			{
			lexer->consumeToken();
#ifdef SCREET_WOOSH
			const char* wooshObject = getenv("SELF");
			if (wooshObject == NULL)
				return ScreetResult(ErrorType, "##!! SELF wasn't defined. !!##");
			return ScreetResult(ObjectType,
			                    ScreetString("##0x") + ScreetString(wooshObject));
#else
			/***/
			return ScreetResult(ErrorType, "##!! \"this\" is not yet implemented. !!##");
#endif
			}
			break;

		case NilToken:
		case FalseToken:
			lexer->consumeToken();
			return ScreetResult();

		case TrueToken:
			lexer->consumeToken();
			return ScreetResult("true");

		case NumberToken:
			{
			double number;
			sscanf(lexer->nextToken().text.c_str(), "%lg", &number);
			lexer->consumeToken();
			return ScreetResult(number);
			}

		case StringLiteralToken:
			{
			ScreetString literal = lexer->nextToken().text;
			lexer->consumeToken();
			return ScreetResult(literal.substr(1, literal.length() - 2));
			}

		case IDToken:
			{
			ScreetString variableName = lexer->nextToken().text;
			lexer->consumeToken();
			if (context)
				return ScreetResult(context->valueFor(variableName));
			else {
				ScreetString errorStr = "##!! There is no context for variables, ";
				errorStr += " so \"";
				errorStr += variableName;
				errorStr += "\" isn't defined. !!##";
				return ScreetResult(ErrorType, errorStr);
				}
			}
			break;

		case OpenParenToken:
			{
			lexer->consumeToken();
			ScreetResult result = parseExpression();
			if (lexer->nextToken().type != CloseParenToken)
				result.error("##!! A \"(\" isn't matched by a \")\". !!##");
			else
				lexer->consumeToken();
			return result;
			}
			break;

		default:
			return ScreetResult(ErrorType, "##!! Some kind of error... !!##");
			break;
		}
}


ScreetArgumentList* ScreetParser::parseArguments()
{
	ScreetArgumentList* args = new ScreetArgumentList();

	ScreetTokenType tokenType = lexer->nextToken().type;
	while (tokenType != CloseParenToken && tokenType != CloseIndexToken) {
		// parse the next argument
		ScreetResult arg = parseExpression();
		if (arg.type == ErrorType) {
			args->error(arg.asString());
			return args;
			}

		// add it to the args
		args->addArgument(arg.asString());

		// the next token must be either a comma, or the end of the args
		tokenType = lexer->nextToken().type;
		if (tokenType == CommaToken)
			lexer->consumeToken();
		else if (tokenType != CloseParenToken && tokenType != CloseIndexToken) {
			args->error("##!! A list of arguments doesn't end, or is missing a comma. !!##");
			return args;
			}
		}

	return args;
}


ScreetArgumentList* ScreetParser::parseCommandArguments()
{
	ScreetArgumentList* args = new ScreetArgumentList();

	while (true) {
		ScreetTokenType tokenType = lexer->nextToken().type;
		if (tokenType == SemicolonToken ||
		    tokenType == CloseParenToken ||
		    tokenType == CloseIndexToken ||
		    tokenType == EOFToken)
			break;

		// parse the next argument
		ScreetResult arg = parseExpression();
		if (arg.type == ErrorType) {
			args->error(arg.asString());
			return args;
			}

		// add it to the args
		args->addArgument(arg.asString());
		}

	return args;
}


ScreetResult ScreetParser::callMethod(ScreetResult value,
                                      ScreetString methodName,
                                      ScreetArgumentList* arguments)
{
	if (!isEvaluating)
		value = ScreetResult();

	else {
		switch (value.type) {
			case ErrorType:
				// leave "value" as the error
				break;

			case NumberType:
				value = callNumberMethod(value.number, methodName, arguments);
				break;

			case StringType:
				value = callStringMethod(value.string, methodName, arguments);
				break;

#ifdef SCREET_WOOSH
			case ObjectType:
				value = screetMethodCall(value.string, methodName, arguments);
				break;
#else
			case ObjectType:
				value = callMethod(value.object, methodName, arguments);
				break;
#endif
			}
		}

	delete arguments;
	return value;
}


ScreetResult ScreetParser::callMethod(void* object, ScreetString methodName,
                                      ScreetArgumentList* args)
{
	// prepare the arguments vector
	int numArgs = args->numArguments();
	ScreetString argsVector[numArgs];
	args->buildVector(argsVector);

	// make the call on the object
	ScreetString resultStr =
		screetMethodCall(object, methodName, numArgs, argsVector);

	ScreetResult result(resultStr);
	return result;
}


ScreetResult ScreetParser::callMethod(ScreetResult object,
                                      const char* methodName,
                                      ScreetResult arg)
{
	ScreetArgumentList* args = new ScreetArgumentList();
	args->addArgument(arg.asString());
	return callMethod(object, ScreetString(methodName), args);
}


ScreetResult ScreetParser::callMethod(ScreetResult object,
                                      const char* methodName)
{
	return callMethod(object, ScreetString(methodName),
	                  new ScreetArgumentList());
}


ScreetResult ScreetParser::callNumberMethod(double number,
                                            ScreetString methodName,
                                            ScreetArgumentList* arguments)
{
	/***/
	return
		ScreetResult(ErrorType,
		             "##!! Non-operator methods on numbers are not yet implemented. !!##");
}


ScreetResult ScreetParser::callStringMethod(ScreetString str,
                                            ScreetString methodName,
                                            ScreetArgumentList* arguments)
{
	ScreetResult result;

	int numArguments = arguments->numArguments();

	if (methodName == "length" && numArguments == 0) {
		result.setTo(str.length());
		}

	else if (methodName == "substr" && (numArguments == 1 || numArguments == 2)) {
		// get the arguments as numbers
		ScreetString args[numArguments];
		arguments->buildVector(args);
		const char* notANumError = "##!! The argument(s) to substr() must be numbers. !!##";
		unsigned int firstChar;
		if (!args[0].isNumber())
			result.error(notANumError);
		else
			firstChar = (unsigned int) (ScreetResult(args[0]).number);
		unsigned int numChars;
		if (result.type != ErrorType && numArguments == 2) {
			if (!args[1].isNumber())
				result.error(notANumError);
			else
				numChars = (unsigned int) (ScreetResult(args[1]).number);
			}

		// get the result
		if (result.type != ErrorType)
			result.setTo(StringType, str.substr(firstChar, numChars));
		}

	else {
		ScreetString errStr =
			"##!! There isn't a method for strings called \"";
		errStr += methodName;
		errStr += "\" with ";
		errStr += ScreetString(numArguments) + " arguments. !!##";
		result.error(errStr);
		}

	return result;
}


