/* ScreetParser.h -  */
/* Copyright 2002 Steve Folta */
/* See the "License" file */

#ifndef _ScreetParser_h_
#define _ScreetParser_h_

#include "ScreetString.h"

class ScreetLexer;
class ScreetContext;
class ScreetArgumentList;


enum ScreetResultType {
	StringType,
	NumberType,
	ObjectType,
	ErrorType
};


struct ScreetResult {
	ScreetResultType	type;
	ScreetString    	string;		// would be part of the union, but C++ doesn't allow it
	union {
		void*         	object;
		double        	number;
	};

	ScreetResult(const ScreetResult& other);
	ScreetResult(ScreetString stringIn);
	ScreetResult(void* objectIn) : type(ObjectType), object(objectIn) {}
	ScreetResult(double numberIn) : type(NumberType), number(numberIn) {}
	ScreetResult(ScreetResultType typeIn, ScreetString stringIn)
		: type(typeIn), string(stringIn) {}
	ScreetResult() : type(StringType), string("") {}

	void         	setTo(ScreetResultType newType, ScreetString newString)
	             		{ type = newType; string = newString; }
	void         	setTo(void* objectIn) { type = ObjectType; object = objectIn; }
	void         	setTo(double numberIn) { type = NumberType; number = numberIn; }
	void         	error(ScreetString errorString)
	             		{ type = ErrorType; string = errorString; }
	ScreetResult&	operator=(const ScreetResult& other);

	ScreetString	asString();
	bool        	conditionIsTrue();
};


class ScreetParser {
public:
	ScreetParser(ScreetString text, ScreetContext* context);
	~ScreetParser();

	ScreetString	parse();

protected:
	ScreetLexer*  	lexer;
	ScreetContext*	context;
	bool          	isEvaluating;

	ScreetResult	parseStatement();
	ScreetResult	parseCompoundStatement();
	ScreetResult	parseIfStatement();
	ScreetResult	parseWhileStatement();
	ScreetResult	parseDoStatement();
	ScreetResult	parseLoopStatement();
	ScreetResult	parseBreakStatement();
	ScreetResult	parseContinueStatement();
	ScreetResult	parseReturnStatement();
	ScreetResult	parseExpressionStatement();
	ScreetResult	parseExpression();
	ScreetResult	parseAssignmentExpression();
	ScreetResult	parseLogicalOrExpression();
	ScreetResult	parseLogicalAndExpression();
	ScreetResult	parseInclusiveOrExpression();
	ScreetResult	parseExclusiveOrExpression();
	ScreetResult	parseAndExpression();
	ScreetResult	parseEqualityExpression();
	ScreetResult	parseAdditiveExpression();
	ScreetResult	parseMultiplicativeExpression();
	ScreetResult	parseUnaryExpression();
	ScreetResult	parsePostfixExpression();
	ScreetResult	parsePrimaryExpression();

	ScreetArgumentList*	parseArguments();
	ScreetArgumentList*	parseCommandArguments();
	ScreetResult       	callMethod(ScreetResult object, ScreetString methodName,
	                   	           ScreetArgumentList* args);
	ScreetResult       	callMethod(void* object, ScreetString methodName,
	                   	           ScreetArgumentList* args);
	ScreetResult       	callMethod(ScreetResult object, const char* methodName,
	                   	           ScreetResult arg);
	ScreetResult       	callMethod(ScreetResult object, const char* methodName);
	ScreetResult       	callNumberMethod(double number, ScreetString methodName,
	                   	                 ScreetArgumentList* arguments);
	ScreetResult       	callStringMethod(ScreetString str, ScreetString methodName,
	                   	                 ScreetArgumentList* arguments);
};


#endif
