/* ScreetString.cpp - String object in Screet */
/* Copyright 2002 Steve Folta */
/* See the "License" file */

#include "ScreetString.h"
#include "ScreetStringStorage.h"
#include <string.h>
#include <stdio.h>		// just for sprintf()


ScreetString::ScreetString(const char* str)
	: storage(NULL)
{
	start = str;
	stopper = str + strlen(str);
}


ScreetString::ScreetString(const char* str, unsigned int numChars,
                           ScreetStringStorage* storageIn)
	: start(str), stopper(str + numChars), storage(storageIn)
{
	if (storage)
		storage->addReference();
}


ScreetString::ScreetString(const ScreetString& str)
	: start(str.start), stopper(str.stopper), storage(str.storage)
{
	if (storage)
		storage->addReference();
}


ScreetString::ScreetString(long int number, const char* format)
{
	if (format == NULL)
		format = "%ld";

	char numberStr[64];
	sprintf(numberStr, format, number);

	storage = new ScreetStringStorage(numberStr, strlen(numberStr));
	start = storage->start();
	stopper = storage->end();
}


ScreetString::~ScreetString()
{
	if (storage)
		storage->removeReference();
}


ScreetString& ScreetString::operator=(const ScreetString& other)
{
	if (storage)
		storage->removeReference();

	start = other.start;
	stopper = other.stopper;
	storage = other.storage;
	if (storage)
		storage->addReference();

	return *this;
}


ScreetString& ScreetString::operator+=(const ScreetString& other)
{
	// optimize adding to the empty string
	if (start >= stopper)
		return *this = other;

	// optimize adding characters that already exist at the end
	if (other.start == stopper) {
		stopper = other.stopper;
		return *this;
		}

	ScreetStringStorage* newStorage =
		new ScreetStringStorage(*this, other);
	if (storage)
		storage->removeReference();
	storage = newStorage;
	start = newStorage->start();
	stopper = newStorage->end();

	return *this;
}


ScreetString ScreetString::operator+(const ScreetString& other)
{
	ScreetStringStorage* newStorage =
		new ScreetStringStorage(*this, other);
	return ScreetString(newStorage->start(), newStorage->length(), newStorage);
}


ScreetString ScreetString::substr(unsigned int firstChar, unsigned int numChars) const
{
	unsigned int len = stopper - start;
	if (firstChar > len)
		firstChar = len;
	if (numChars > len - firstChar)
		numChars = len - firstChar;

	return ScreetString(start + firstChar, numChars, storage);
}



bool ScreetString::operator==(const ScreetString& other) const
{
	const char* thisP = start;
	const char* otherP = other.start;
	const char* otherStopper = other.stopper;
	while (true) {
		// check for end
		if (thisP >= stopper)
			return (otherP == otherStopper);
		else if (otherP >= otherStopper)
			return false;

		// compare
		if (*thisP++ != *otherP++)
			return false;
		}
}


bool ScreetString::operator!=(const ScreetString& other) const
{
	const char* thisP = start;
	const char* otherP = other.start;
	const char* otherStopper = other.stopper;
	while (true) {
		// check for end
		if (thisP >= stopper)
			return (otherP != otherStopper);
		else if (otherP >= otherStopper)
			return true;

		// compare
		if (*thisP++ == *otherP++)
			return false;
		}
}


bool ScreetString::startsWith(const ScreetString& other) const
{
	if (other.length() > length())
		return false;

	const char* p1 = start;
	const char* p2 = other.start;
	while (p2 < other.stopper) {
		if (*p1++ != *p2++)
			return false;
		}
	return true;
}


bool ScreetString::startsWith(const char* other) const
{
	const char* p1 = start;
	const char* p2 = other;
	while (*p2) {
		if (p1 >= stopper || *p1++ != *p2++)
			return false;
		}
	return true;
}


bool ScreetString::isNumber() const
{
	const char* p = start;

	// the empty string is not a number
	if (p >= stopper)
		return false;

	// can optionally begin with '-'
	if (p < stopper && *p == '-')
		++p;

	// then there must be numbers...
	char c;
	while (p < stopper) {
		c = *p++;
		if (c < '0' || c > '9')
			return false;
		}

	// It's at least *starting* to be a number.
	// After this point, we just match what we can.  If we don't match
	// the whole string by the end of the function, it's not a number.

	// then there can be a period, optionally followed by more numbers
	if (p < stopper && *p == '.') {
		++p;
		while (p < stopper) {
			char c = *p++;
			if (c < '0' || c > '9')
				break;
			}
		}

	// then there can be an 'e' followed by more numbers
	if (p < stopper && *p == 'e') {
		++p;
		// must have at least one digit after the 'e'
		if (p >= stopper)
			return false;
		c = *p++;
		if (c < '0' || c > '9')
			return false;
		// then more digits
		while (p < stopper) {
			c = *p++;
			if (c < '0' || c > '9')
				break;
			}
		}

	// okay, we've matched a number.  Did we match the whole string?
	return (p >= stopper);
}


const char* ScreetString::c_str()
{
	if (storage == NULL || storage->end() != stopper) {
		unsigned long len = length();
		ScreetStringStorage* newStorage =
			new ScreetStringStorage(start, len);
		if (storage)
			storage->removeReference();
		storage = newStorage;
		start = storage->start();
		stopper = start + len;
		}

	return start;
}


void ScreetString::setTo(const char* newStart, const char* newStopper)
{
	if (storage)
		storage->removeReference();
	start = newStart;
	stopper = newStopper;
	storage = NULL;
}


void ScreetString::detach()
{
	if (storage && storage->referenceCount() == 1) {
		// we're already detached
		return;
		}

	ScreetStringStorage* newStorage = new ScreetStringStorage(start, length());
	if (storage)
		storage->removeReference();
	storage = newStorage;
	start = storage->start();
	stopper = storage->end();
}


void ScreetString::semiDetach()
{
	if (storage) {
		// we're already semi-detached
		return;
		}

	storage = new ScreetStringStorage(start, length());
	start = storage->start();
	stopper = storage->end();
}


