/* ScreetTo.cpp - Command line to talk to Screet on a remote port. */
/* Copyright 2002 Steve Folta */
/* See the "License" file */

#include <stdio.h>		// must come before readline stuff
#include <readline/readline.h>
#include <readline/history.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <string>


// tweex
#define kBufferChunkSize	512


class Error {
public:
	Error(string messageIn) : message(messageIn) {}

	string	message;
};


class TCPConnection {
public:
	TCPConnection(const char* hostName, int port);
	~TCPConnection();

	void 	send(const char* message);
	char*	receive();	// Makes the result.  NULL if the connection closed.

protected:
	int	connectionSocket;
};


int main(int argc, char* argv[])
{
	if (argc < 2 || argc > 3) {
		printf("Usage: screet-to ip-address [port]\n");
		return 1;
		}

	const char* screetServerName = argv[1];
	int port = 1998;
	if (argc >= 3)
		port = atoi(argv[2]);

	try {
		TCPConnection connection(screetServerName, port);

		while (true) {
			char* line = readline("screet> ");
			if (line == NULL)
				break;
			if (line[0] == 0) {
				free(line);
				break;
				}

			connection.send(line);
			char* response = connection.receive();
			printf("%s\n\n", response);

			add_history(line);

			free(line);
			free(response);
			}
		}
	catch (Error& error) {
		fprintf(stderr, "%s\n", error.message.c_str());
		}

	return 0;
}



TCPConnection::TCPConnection(const char* hostName, int port)
	: connectionSocket(-1)
{
	// get the host address
	struct hostent* hostAddr = gethostbyname(hostName);
	if (hostAddr == NULL)
		throw Error("Couldn't find the Screet server.");

	// create the socket
	connectionSocket = socket(AF_INET, SOCK_STREAM, 0);
	if (connectionSocket == -1)
		throw Error("Couldn't open a socket.");

	// bind
	struct sockaddr_in address;
	memset(&address, 0, sizeof(address));
	address.sin_family = AF_INET;
	address.sin_port = htons(0);	// use any port
	address.sin_addr.s_addr = INADDR_ANY;
	int result =
		bind(connectionSocket, (struct sockaddr*) &address, sizeof(struct sockaddr));
	if (result == -1) {
		close(connectionSocket);
		connectionSocket = -1;
		throw Error("Couldn't bind the socket.");
		}

	// connect
	struct sockaddr_in destAddr;
	memset(&destAddr, 0, sizeof(destAddr));
	destAddr.sin_family = AF_INET;
	destAddr.sin_port = htons(port);
	destAddr.sin_addr = *((struct in_addr *) hostAddr->h_addr);
	result =
		connect(connectionSocket, (struct sockaddr*) &destAddr, sizeof(destAddr));
	if (result == -1) {
		close(connectionSocket);
		connectionSocket = -1;
		throw Error("Couldn't connect to the Screet server.");
		}
}


TCPConnection::~TCPConnection()
{
	if (connectionSocket != -1)
		close(connectionSocket);
}


void TCPConnection::send(const char* message)
{
	int bytesLeft = strlen(message) + 1;
	while (bytesLeft > 0) {
		int bytesSent = ::send(connectionSocket, message, bytesLeft, 0);
		if (bytesSent < 0)
			throw Error("Error while sending.");
		message += bytesSent;
		bytesLeft -= bytesSent;
		}
}


char* TCPConnection::receive()
{
	// read from the connection
	long bufferSize = kBufferChunkSize;
	char* buffer = (char*) malloc(bufferSize);
	int bytesLeftInBuffer = bufferSize;
	int totalBytesRead = 0;
	while (true) {
		// read as much as we can
		int bytesRead =
			recv(connectionSocket, &buffer[totalBytesRead], bytesLeftInBuffer, 0);
		if (bytesRead == 0) {
			// the connection was closed, before we received a complete expression
			close(connectionSocket);
			free(buffer);
			return NULL;
			}
		totalBytesRead += bytesRead;
		bytesLeftInBuffer -= bytesRead;

		// are we done?
		if (buffer[totalBytesRead - 1] == 0)
			break;

		// did we run out of buffer space?
		if (bytesLeftInBuffer <= 0) {
			// add another chunk
			bufferSize += kBufferChunkSize;
			buffer = (char*) realloc(buffer, bufferSize);
			bytesLeftInBuffer += kBufferChunkSize;
			}
		}

	if (totalBytesRead == 0) {
		free(buffer);
		return false;
		}

	return buffer;
}


